/****************************************************************************
 *                                                                          *
 * Duration of voiced regions in F0 contours (dur.c)                        *
 *                                                                          *
 * Implemented by Paul Bagshaw, Centre for Speech Technology Research,      *
 * University of Edinburgh, 80 South Bridge, Edinburgh EH1 1HN.             *
 *                                                                          *
 * 1st - 15th August 1991                                                   *
 *                                                                          *
 ****************************************************************************/

#include <math.h>
#include <stdio.h>
#include <values.h>
#include "cleanup.h"

typedef enum {
  CANT_OPEN, CANT_READ, MISUSE, NOT_XMG, SEG_ONLY
  } error_types;

typedef struct {
  int count;
  float voiced_duration, unvoiced_duration;
  double sum_err, sum_err_sqr;
} STATISTICS_;

static char program[20] = "dur";

void main (argc, argv)
     int argc;
     char *argv[];
{

  void error (), initialise_statistics (), read_contour ();

  int n, in_int, sample_freq = 20;
  char in_str[100], in_char = '\0';
  FILE *infile;
  STATISTICS_ stats;
  float time, f0, prev_time;
  double f0_err;

  /* parse parameters */
  if (argc < 2) error (MISUSE);
  initialise_statistics (&stats);
  for (n = 1; n < argc; n++) {
    if ((infile = fopen (argv[n], "r")) == NULL) error (CANT_OPEN, argv[n]);
    /* read Xmg fileheaders */
    if (fscanf (infile, "%s", in_str) != 1) error (CANT_READ, argv[n]);
    if (strcmp (in_str, "XAO1")) error (NOT_XMG, argv[n]);
    if (!fread ((char *) &in_char, sizeof (char), 1, infile))
      error (CANT_READ, argv[n]);
    while (!feof (infile) && in_char != '\14') {
      if (in_char == '#' || in_char == '\n') {
	while (in_char != '\n')
	  if (!fread ((char *) &in_char, sizeof (char), 1, infile))
	    error (CANT_READ, argv[n]);
	if (!fread ((char *) &in_char, sizeof (char), 1, infile))
	  error (CANT_READ, argv[n]);
	continue;
      }
      fseek (infile, (long) -sizeof (char), 1);
      if (fscanf (infile, "%s", in_str) != 1) error (CANT_READ, argv[n]);
      if (!strcmp (in_str, "LineType")) {
	if (fscanf (infile, "%s", in_str) != 1) error (CANT_READ, argv[n]);
	if (strcmp (in_str, "segments")) error (SEG_ONLY, argv[n]);
      }
      else if (!strcmp (in_str, "Freq")) {
	if (fscanf (infile, "%d", &in_int) != 1) error (CANT_READ, argv[n]);
	sample_freq = in_int * 1000;
      }
      while (in_char != '\n')
	if (!fread ((char *) &in_char, sizeof (char), 1, infile))
	  error (CANT_READ, argv[n]);
      if (!fread ((char *) &in_char, sizeof (char), 1, infile))
	error (CANT_READ, argv[n]);
    }
    /* ignore Xmg fileheaders */
    rewind(infile);
    in_char = '\0';
    while (!feof (infile) && in_char != '\14')
      if (!fread ((char *) &in_char, sizeof (char), 1, infile))
        error (CANT_READ, argv[n]);
    if (!fread ((char *) &in_char, sizeof (char), 1, infile))
      error (CANT_READ, argv[n]);
    if (in_char != '\n') error (CANT_READ, argv[n]);
    /* modify durations */
    read_contour (infile, &prev_time, &f0);
    if (f0 != 0.0) {
      f0_err = sample_freq / ((sample_freq * sample_freq) / (f0 * f0) - 1);
      stats.sum_err += f0_err;
      stats.sum_err_sqr += f0_err * f0_err;
      stats.count++;
    }
    while (!feof (infile)) {
      read_contour (infile, &time, &f0);
      if (f0 == 0.0)
	stats.unvoiced_duration += time - prev_time;
      else {
	stats.voiced_duration += time - prev_time;
	f0_err = sample_freq / ((sample_freq * sample_freq) / (f0 * f0) - 1);
	stats.sum_err += f0_err;
	stats.sum_err_sqr += f0_err * f0_err;
	stats.count++;
      }
      prev_time = time;
    }
  }
  /* display durations */
  printf (" Unvoiced regions duration : %f ms\n", stats.unvoiced_duration);
  printf (" Voiced regions duration : %f ms\n", stats.voiced_duration);
  printf (" Mean F0 Error: %f\n", stats.sum_err / stats.count);
  printf (" Population Standard Deviation of F0 Error: %f\n",
	  sqrt ((stats.sum_err_sqr -
		 stats.sum_err * stats.sum_err / stats.count) /
		(stats.count - 1)));
  exit (0);

}

void error (flag, name)
     error_types flag;
     char *name;
{

  if (flag != MISUSE)
    fprintf (stderr, "%s: ", program);
  switch (flag) {
  case CANT_OPEN:
    fprintf (stderr, "cannot open file named %s", name);
    break;
  case CANT_READ:
    fprintf (stderr, "cannot read data from %s", name);
    break;
  case MISUSE:
    fprintf (stderr, "usage: %s F0_file ...", program);
    break;
  case NOT_XMG:
    fprintf (stderr, "input file %s not in Xmg format", name);
    break;
  case SEG_ONLY:
    fprintf (stderr, "input file %s doesn't have LineType 'segments'", name);
    break;
  default:
    fprintf (stderr, "undefined error, %d occured", flag);
    break;
  }
  fprintf (stderr, "\n");
  exit (-1);

}

void initialise_statistics (p_stats)
     STATISTICS_ *p_stats;
{

  p_stats->count = 0;
  p_stats->voiced_duration = 0.0;
  p_stats->unvoiced_duration = 0.0;
  p_stats->sum_err = 0.0;
  p_stats->sum_err_sqr = 0.0;

}

void read_contour (f0_file, p_time, p_f0)
     FILE *f0_file;
     float *p_time, *p_f0;
{

  char i;
  long pos;

  if (!fread ((char *) &i, sizeof (char), 1, f0_file)) error (CANT_READ, NULL);
  if (i == '=' || i == '\n') {
    if (!fread ((char *) &i, sizeof (char), 1, f0_file) || i != '\n')
      error (CANT_READ, NULL);
    pos = ftell (f0_file);
    if (fscanf (f0_file, "%f ", p_time) != 1) error (CANT_READ, NULL);
    fseek (f0_file, pos, 0);
    *p_f0 = 0.0;
    return;
  }
  fseek (f0_file, (long) -sizeof (char), 1);
  if (fscanf (f0_file, "%f %f\n", p_time, p_f0) != 2) error (CANT_READ, NULL);
  return;

}
